import rsa
from base64 import b64encode, b64decode

from .exceptions import *
from .steam_api_web import *
from .steam_session import SteamSession

import hmac
import base64
import struct
import hashlib

from .steam_api_mobile import *

class LoginConfirmType:
    none = 1
    email = 2
    mobile = 3

class SteamMobile(SteamSession):

    def __init__(self, account_name: str, password: str) -> None:

        self.__client_id = None
        self.__request_id = None
        self.__code_type = None

        self.device_id = generate_device_id()
        self.shared_secret = None
        self.serial_number = None
        self.revocation_code = None
        self.uri = None
        self.token_gid = None
        self.identity_secret = None
        self.secret_1 = None

        super().__init__(account_name=account_name, password=password)

    def load(self, data: dict):

        if not data: return

        self.steamid = data.get('steamid')
        self.access_token = data.get('access_token')
        self.refresh_token = data.get('refresh_token')
        self.session_id = data.get('session_id')

        for c in data.get('cookies', []): self.session.cookies.set(c['name'], c['value'], domain=c['domain'], path=c['path'], secure=c['secure'], expires=c['expires'])

    def load_mobile(self, data: dict):

        if not data: return

        self.device_id = data.get('device_id')

        self.shared_secret = data.get('shared_secret')
        self.serial_number = data.get('serial_number')
        self.revocation_code = data.get('revocation_code')
        self.uri = data.get('uri')
        self.token_gid = data.get('token_gid')
        self.identity_secret = data.get('identity_secret')
        self.secret_1 = data.get('secret_1')

    def login(self) -> LoginConfirmType:

        key, encryption_timestamp = getPasswordRSAPublicKey(self.session, self.account_name)

        encrypted_pass = b64encode(rsa.encrypt(self._password.encode(), key)).decode()

        res = beginAuthSessionViaCredentials(self.session, self.account_name, encrypted_pass, encryption_timestamp)

        if not res.get('steamid') or not res.get('client_id') or not res.get('request_id'): raise InvalidCredentials()

        self.steamid = res.get('steamid')
        self.__client_id = res.get('client_id')
        self.__request_id = res.get('request_id')

        # TODO: Check allowed_confirmations and make sure if it's without steamguard, email steamguard or mobile steamguard
        allowed_confirmations = res.get('allowed_confirmations') # Mobile -> 'allowed_confirmations': [{'confirmation_type': 3}], Email -> 2

        self.__code_type = int(allowed_confirmations[0].get('confirmation_type'))

        return self.__code_type #, allowed_confirmations[0].get('associated_message')

    def confirm_login(self, steam_guard_code: str):

        updateAuthSessionWithSteamGuardCode(self.session, self.__client_id, self.steamid, steam_guard_code, self.__code_type) # INFO: We're not getting anything interesting from here

        res = pollAuthSessionStatus(self.session, self.__client_id, self.__request_id)

        if not res.get('account_name') or not res.get('access_token') or not res.get('refresh_token'): raise InvalidSteamGuardCode()

        self.access_token = res.get('access_token')
        self.refresh_token = res.get('refresh_token')

        finalizelogin(self.session, self.refresh_token)
        #self.session_id = self.session.cookies.get('sessionid', domain='steamcommunity.com') # TODO: Is session_id really necessary to be set? Can't we generate one?
    
    def get_steampowered(self): self.session.get('https://store.steampowered.com/')
    def get_steamcommunity(self): self.session.get('https://steamcommunity.com/')

    def add_phone_number(self, country_code: str, phone_number: str): # 48 - Poland :)

        res = phone_add_ajaxop(self.session, 'get_phone_number', f'+{country_code} {phone_number}')
        # => {'success': False, 'showResend': False, 'state': False, 'errorText': 'Your account already has a phone number attached to it.', 'token': '0'}
        # => {'success': True, 'showResend': False, 'state': 'email_verification', 'errorText': '', 'token': '0', 'phoneNumber': '+XX123456789'}

        if not res.get('success'):
            if res.get('errorText') == 'Your account already has a phone number attached to it.': raise AlreadyHasAPhoneNumber()
            raise UnknownException(res) # INFO: Well, I don't know every error :)

        return res

    def add_phone_number_email_verified(self):

        res = phone_add_ajaxop(self.session, 'email_verification')
        # => {'success': True, 'showResend': False, 'state': 'get_sms_code', 'errorText': '', 'token': '0', 'inputSize': '20', 'maxLength': '5'}
        #print('email verified ->', res)
        if res.get('state') == 'retry_email_verification': raise EmailNotVerified()
        if not res.get('success'): raise UnknownException(res)
        return res

    def add_phone_number_sms_code(self, code: str):

        res = phone_add_ajaxop(self.session, 'get_sms_code', code)
        # => {'success': True, 'showResend': False, 'state': 'done', 'errorText': '', 'token': '0', 'vac_policy': 0, 'tos_policy': 2, 'showDone': True, 'maxLength': '5'}
        # => {'success': False, 'showResend': False, 'state': False, 'errorText': 'Bad SMS code, please try again (CASM)', 'token': '0'}
        #print('get_sms_code ->', res)
        if res.get('errorText') == 'Bad SMS code, please try again (CASM)': raise InvalidSMSCode()
        if not res.get('success'): raise UnknownException(res)
        return res

    # WEB SESSION IS SIMULATED UP HERE
    # MOBILE SESSION IS SIMULATED DOWN HERE

    def deactivate_mobile_auth(self, steamguard_scheme: int = 1): # 1 - Return to email codes, 2 - Remove completly
        res = deactivateAuthenticator(self.session, self.revocation_code, self.access_token, steamguard_scheme)
        if not res.get('success'): raise UnknownException(res)
        return res

    def add_mobile_auth(self):
        res = addAuthenticator(self.session, self.access_token, self.steamid, self.device_id, str(self.get_steam_time()))
        status = res.get('status')

        # status codes TODO: Add more
        # 1 - ok
        # 2 - no phone number is on the account
        # 29 Already secured by Mobile Steam Guard

        if status == 29: raise AlreadyHasMobileSteamGuard()
        if not res.get('shared_secret') and not res.get('revocation_code'): raise UnknownException(res)

        self.shared_secret = res.get('shared_secret')
        self.serial_number = res.get('serial_number')
        self.revocation_code = res.get('revocation_code') # INFO: You can't lose this code!!! Save it every time!
        self.uri = res.get('uri')
        self.token_gid = res.get('token_gid')
        self.identity_secret = res.get('identity_secret')
        self.secret_1 = res.get('secret_1')

        return res

    def add_mobile_auth_confirm(self, sms_code: str):

        steam_guard_code = self.generate_steam_guard_code()

        res = finalizeAddAuthenticator(self.session, self.access_token, self.steamid, sms_code, steam_guard_code, str(self.get_steam_time()))

        if not res.get('success') or not res.get('status') == 2: raise UnknownException(res)

        if res.get('status') == 88: raise UnableToGenerateCorrectCodes()
        if res.get('status') == 89: raise InvalidSMSCode()

        return res

    def generate_steam_guard_code(self): return self.generate_steam_guard_code_for_time(self.get_steam_time())

    def generate_steam_guard_code_for_time(self, t: int):
        if not self.shared_secret: return None
        time_buffer = struct.pack('>Q', t // 30)
        time_hmac = hmac.new(base64.b64decode(self.shared_secret), time_buffer, digestmod=hashlib.sha1).digest()
        begin = ord(time_hmac[19:20]) & 0xF
        full_code = struct.unpack('>I', time_hmac[begin:begin + 4])[0] & 0x7FFFFFFF
        chars = '23456789BCDFGHJKMNPQRTVWXY'
        code = ''
        for _ in range(5):
            full_code, i = divmod(full_code, len(chars))
            code += chars[i]
        return code

    def export_mobile(self):
        return {
            'device_id': self.device_id,
            'shared_secret': self.shared_secret,
            'serial_number': self.serial_number,
            'revocation_code': self.revocation_code,
            'uri': self.uri,
            'token_gid': self.token_gid,
            'identity_secret': self.identity_secret,
            'secret_1': self.secret_1
        }

    # ====== #
    # TRADES #
    # ====== #

    def _send_confirmation(self, data_confid: str, nonce: str) -> dict:
        params = self._create_confirmation_params('allow')
        params['op'] = ('allow',)
        params['cid'] = data_confid
        params['ck'] = nonce
        headers = { 'X-Requested-With': 'XMLHttpRequest' }
        return self.session.get(f'https://steamcommunity.com/mobileconf/ajaxop', params=params, headers=headers).json()

    def _generate_confirmation_key(self, identity_secret: str, tag: str, timestamp: int) -> bytes:
        buffer = struct.pack('>Q', timestamp) + tag.encode('ascii')
        return b64encode(hmac.new(b64decode(identity_secret), buffer, digestmod=hashlib.sha1).digest())

    def get_trade_confirmations(self):
        params = self._create_confirmation_params('conf')
        headers = { 'X-Requested-With': 'com.valvesoftware.android.steam.community' }
        return self.session.get(f'https://steamcommunity.com/mobileconf/getlist', params=params, headers=headers)

    def _create_confirmation_params(self, tag_string: str):
        timestamp = int(time.time())
        confirmation_key = self._generate_confirmation_key(self.identity_secret, tag_string, timestamp)
        return {
            'p': self.device_id,
            'a': self.steamid,
            'k': confirmation_key,
            't': timestamp,
            'm': 'android',
            'tag': tag_string
        }

# TODO:
# Add tradeoffers confirmations
# Get inventory
# Get tradeable inventory


import time
import uuid
import requests

def generate_device_id(): return 'android:' + str(uuid.uuid4()) # "android:" + Guid.NewGuid().ToString()

def getSteamQueryTimeDifference(session: requests.Session):

    currentTime = int(time.time())

    params = { 'steamid': '0' }

    url = 'https://api.steampowered.com/ITwoFactorService/QueryTime/v0001'
    res = session.post(url, params=params).json()

    res = res.get('response')

    return int(int(res.get('server_time')) - currentTime)

def addAuthenticator(session: requests.Session, access_token: str, steamid: str, device_id: str, authenticator_time: str = None):

    data = {
        'steamid': steamid,
        'authenticator_type': '1',
        'device_identifier': device_id,
        #'authenticator_time': authenticator_time,
        'sms_phone_id': '1',
        'version': 2
    }

    params = { 'access_token': access_token }

    url = 'https://api.steampowered.com/ITwoFactorService/AddAuthenticator/v1'
    res = session.post(url, params=params, data=data)

    if res.status_code != 200: return None

    res = res.json()
    res = res.get('response')
    return res

def finalizeAddAuthenticator(session: requests.Session, access_token: str, steamid: str, code: str, steam_guard_code: str, authenticator_time: str):

    data = {
        'steamid': steamid,
        'activation_code': code,
        'authenticator_code': steam_guard_code,
        'authenticator_time': authenticator_time,
        #'validate_sms_code': '1'
    }

    params = { 'access_token': access_token }

    url = 'https://api.steampowered.com/ITwoFactorService/FinalizeAddAuthenticator/v1'
    res = session.post(url, params=params, data=data).json()

    res = res.get('response')
    return res

def deactivateAuthenticator(session: requests.Session, revocation_code: str, access_token: str, steamguard_scheme: int = 1): # 1 - Return to email codes, 2 - Remove completley

    data = {
        'revocation_code': revocation_code,
        'revocation_reason': '1',
        'steamguard_scheme': str(steamguard_scheme)
    }

    params = { 'access_token': access_token }

    url = "https://api.steampowered.com/ITwoFactorService/RemoveAuthenticator/v1"
    res = session.post(url, params=params, data=data).json()
    return res.get('response')

def fetchConfirmations(session: requests.Session, params: dict):
    res = session.get('https://steamcommunity.com/mobileconf/getlist', params=params).json()
    return res

def sendConfirmationAjaxop(session: requests.Session, params: dict):
    res = session.get('https://steamcommunity.com/mobileconf/ajaxop', params=params).json()
    return res


import os
import json
import time
import base64
import requests
from datetime import datetime, timezone

from .exceptions import *

USER_AGENT_MOBILE = 'Dalvik/2.1.0 (Linux; U; Android 9; Valve Steam App Version/3)'

class SteamSession:

    def __init__(self, account_name: str, password: str) -> None:

        self.account_name = account_name
        self._password = password

        self.steamid = None
        self.access_token = None
        self.refresh_token = None # INFO: If refresh_token is expired, I guess we've to login again?
        self.session_id = None

        self.__steamTimeDiff = 0
        self.__steamTimeAligned = False

        self.session = requests.Session()
        self.session.headers.update({ 'User-Agent': USER_AGENT_MOBILE })

        self.default_folder = '.th3poli-steamguard'

    def refresh_access_token(self):
        if not self.refresh_token: raise RefreshTokenEmpty()
        if self.is_token_expired(self.refresh_token): raise RefreshTokenExpired()

        data = { 'steamid': self.steamid, 'refresh_token': self.refresh_token }

        res = self.session.post("https://api.steampowered.com/IAuthenticationService/GenerateAccessTokenForApp/v1", data=data)
        res = res.json().get('response')
        self.access_token = res.get('access_token', self.access_token)
        self.refresh_token = res.get('refresh_token', self.refresh_token)
        pre = self.steamid + '%7C%7C'
        self.session.cookies.set('steamLoginSecure', pre + self.access_token, domain='steamcommunity.com')
        self.session.cookies.set('steamLoginSecure', pre + self.access_token, domain='store.steampowered.com')
        self.session.cookies.set('steamRefresh_steam', pre + self.refresh_token, domain='login.steampowered.com')

    def is_token_expired(self, token: str): return datetime.now(timezone.utc).timestamp() > self.get_token_expire_timestamp(token) # TODO: Can't it be just int(time.time())

    def get_token_expire_timestamp(self, token: str):
        if not token: return 0

        token_components = token.split('.')
        base64_str = token_components[1].replace('-', '+').replace('_', '/')
        if len(base64_str) % 4 != 0: base64_str += '=' * (4 - len(base64_str) % 4)

        payload_bytes = base64.b64decode(base64_str)
        return json.loads(payload_bytes.decode('utf-8')).get('exp')

    def get_steam_time(self):
        if not self.__steamTimeAligned: self.align_time()
        return int(time.time() + self.__steamTimeDiff)

    def align_time(self):
        currentTime = int(time.time())

        params = { 'steamid': '0' }
        res = self.session.post('https://api.steampowered.com/ITwoFactorService/QueryTime/v0001', params=params).json()
        res = res.get('response')

        self.__steamTimeAligned = True
        self.__steamTimeDiff = int(int(res.get('server_time')) - currentTime)

    def export(self):
        return {
            'steamid': self.steamid,
            'access_token': self.access_token,
            'refresh_token': self.refresh_token,
            'session_id': self.session_id,
            'account_name': self.account_name,
            'cookies': self.export_cookies_dict()
        }

    def export_cookies_dict(self):
        return [{ 'name': c.name, 'value': c.value, 'domain': c.domain, 'path': c.path, 'secure': c.secure, 'expires': c.expires } for c in self.session.cookies]

    def export_cookies(self):
        return [c for c in self.session.cookies]

    def save_exported_data(self, data: dict, path: str):
        os.makedirs(self.default_folder, exist_ok=True)
        path = os.path.join(self.default_folder, path)
        with open(path, 'w', encoding='utf-8') as file: file.write(json.dumps(data))

    def load_exported_data(self, path: str, default = None):
        os.makedirs(self.default_folder, exist_ok=True)
        path = os.path.join(self.default_folder, path)
        if not os.path.isfile(path): return default
        with open(path, 'r', encoding='utf-8') as file: return json.loads(file.read())



from setuptools import setup

setup(
    name='steamguard',
    version='1.0.4',
    description='Simple python module to add a phone number to your steam account and enable mobile auth. Also generate steam guard codes.',
    license='MIT',
    author='th3poli',
    author_email='',
    keywords=['steam', 'mobile', 'auth', 'guard', 'steamcommunity'],
    url='https://github.com/th3poli/steamguard',
    install_requires=['requests', 'rsa']
)

# python setup.py sdist bdist_wheel
# pip install ./dist/steamguard-1.0.4.tar.gz --no-cache-dir
# python -m twine upload .\dist\*



import rsa
import requests
from bs4 import BeautifulSoup

def getPasswordRSAPublicKey(session: requests.Session, account_name: str):
    url = 'https://api.steampowered.com/IAuthenticationService/GetPasswordRSAPublicKey/v1'
    res = session.get(url, params={ 'account_name': account_name }).json()
    res = res.get('response')
    public_key = rsa.PublicKey(int(res['publickey_mod'], 16), int(res['publickey_exp'], 16))
    return public_key, int(res['timestamp'])

def beginAuthSessionViaCredentials(session: requests.Session, account_name: str, encrypted_password: str, encryption_timestamp: int):

    params = {
        'account_name': account_name,
        'encrypted_password': encrypted_password,
        'encryption_timestamp': encryption_timestamp,
        'persistence': '1',
        #'device_friendly_name': '',
        'remember_login': 'true',
        'platform_type': '3',
        #'guard_data': '',
        #'language': '',
        #'device_details': session.headers.get('User-Agent')
    }

    url = 'https://api.steampowered.com/IAuthenticationService/BeginAuthSessionViaCredentials/v1'
    res = session.post(url, params=params).json()
    res = res.get('response')
    return res

    client_id = res.get('client_id')
    request_id = res.get('request_id')
    allowed_confirmations = res.get('allowed_confirmations')
    steamid = res.get('steamid')

    if allowed_confirmations and steamid:

        print('steamid ->', steamid)
        print('allowed confirmations ->', allowed_confirmations)
        
        for allowed_conf in allowed_confirmations:
            confirmation_type = allowed_conf.get('confirmation_type')
            associated_message = allowed_conf.get('associated_message')
            if confirmation_type == 2: # email guard
                print('Get steam guard from ->', associated_message)

    return client_id, request_id, steamid

def updateAuthSessionWithSteamGuardCode(session: requests.Session, client_id: str, steamid: str, code: str, code_type: str = '2'):

    params = {
        'client_id': client_id,
        'steamid': steamid,
        'code': code,
        'code_type': code_type
    }

    url = 'https://api.steampowered.com/IAuthenticationService/UpdateAuthSessionWithSteamGuardCode/v1'
    res = session.post(url, params=params)
    return res

def pollAuthSessionStatus(session: requests.Session, client_id: str, request_id: str) -> dict:

    params = {
        'client_id': client_id,
        'request_id': request_id
    }

    url = 'https://api.steampowered.com/IAuthenticationService/PollAuthSessionStatus/v1'
    res = session.post(url, params=params).json()
    return res.get('response')

def finalizelogin(session: requests.Session, refresh_token: str):

    params = {
        'nonce': refresh_token,
        'sessionid': session.cookies.get(name='sessionid', domain='steamcommunity.com'),
        'redir': 'https://steamcommunity.com/login/home/?goto='
    }

    url = 'https://login.steampowered.com/jwt/finalizelogin'
    res = session.post(url, data=params)

    res = res.json()

    steamID = res.get('steamID')
    transfer_info = res.get('transfer_info')

    for transfer in transfer_info:
        transfer_url = transfer.get('url')

        if not transfer_url in ['https://store.steampowered.com/login/settoken', 'https://steamcommunity.com/login/settoken']: continue

        transfer_params = transfer.get('params')
        transfer_params['steamID'] = steamID

        session.post(transfer_url, params=transfer_params)

    return res

'''
get_phone_number -> {
    "success": true,
    "showResend": false,
    "state": "email_verification",
    "errorText": "",
    "token": "0",
    "phoneNumber": "+XX123456789"
}
'''

def phone_add_ajaxop(session: requests.Session, op: str = 'get_phone_number' or 'email_verification' or 'get_sms_code', input_: str = ''):

    data = {
        'op': op,
        'input': input_,
        'sessionID': session.cookies.get(name='sessionid', domain='store.steampowered.com'),
        'confirmed': '1',
        'checkfortos': '1',
        'bisediting': '0',
        'token': '0'
    }

    url = 'https://store.steampowered.com/phone/add_ajaxop'
    res = session.post(url, data=data)
    return res.json()

def get_tradelink(session: requests.Session):
    res = session.get('https://steamcommunity.com/my/tradeoffers/privacy')
    soup = BeautifulSoup(res.text, 'html.parser')
    element = soup.find(id='trade_offer_access_url')
    tradelink = element.get('value')
    return tradelink

def get_inventory(session: requests.Session, steamid: str): return session.get(f"https://steamcommunity.com/inventory/{steamid}/730/2?l=polish&count=2000")

def get_tradeable_inventory(session: requests.Session, steamid: str):

    inventory = get_inventory(session, steamid).json()

    if not inventory: return []
    if inventory['total_inventory_count'] == 0: return []

    assets = inventory['assets']
    descriptions = inventory['descriptions']

    ready_assets = []
    for asset in assets:
        classid = asset['classid']
        assetid = asset['assetid']

        for desc in descriptions:
            if desc['classid'] == classid and desc['tradable']:
                ready_assets.append({ "tradeable": desc['tradable'], "assetid": assetid, 'tradeoffer_asset': { "appid":730,"contextid":"2","amount":1,"assetid": assetid } })
                break

    return ready_assets

import json
def send_tradeoffer(session: requests.Session, partner_steamid: str, tradeoffer_token: str, items: list, tradeoffermessage: str = ''):

    tradeoffer_headers = { "Referer": "https://steamcommunity.com/tradeoffer/new" }
    assets = []
    for item in items: assets.append(item['tradeoffer_asset'])
    tradeoffer = {
        'sessionid': session.cookies.get(name='sessionid', domain='steamcommunity.com'),
        'serverid': 1,
        'partner': partner_steamid,
        'tradeoffermessage': tradeoffermessage,
        'json_tradeoffer': json.dumps({"newversion": True, "version": 2, "me": { "assets": assets, "currency": [], "ready": False }, "them": { "assets": [], "currency": [], "ready": False }}),
        'captcha': '',
        'trade_offer_create_params': json.dumps({ "trade_offer_access_token": tradeoffer_token })
    }
    return session.post('https://steamcommunity.com/tradeoffer/new/send', data=tradeoffer, headers=tradeoffer_headers)

def accept_tradeoffer(session: requests.Session, tradeofferid: int, partner: int):
    headers = { 'Referer': f'https://steamcommunity.com/tradeoffer/{tradeofferid}' }
    payload = {
        'sessionid': session.cookies.get('sessionid', domain='steamcommunity.com'),
        'serverid': '1', 'tradeofferid': str(tradeofferid), 'partner': str(partner), 'captcha': ''
    }
    return session.post(f'https://steamcommunity.com/tradeoffer/{tradeofferid}/accept', data=payload, headers=headers)

def get_tradeoffer_items(session: requests.Session, tradeofferid: int):
    html = session.get(f'https://steamcommunity.com/tradeoffer/{tradeofferid}').text
    soup = BeautifulSoup(html, 'html.parser')
    items = soup.find_all('a', { 'class': 'inventory_item_link' })
    return items




class UnknownException(Exception): pass

class RefreshTokenEmpty(Exception): pass
class RefreshTokenExpired(Exception): pass

class InvalidCredentials(Exception): pass
class InvalidSteamGuardCode(Exception): pass

class NoAccountName(Exception): pass
class AlreadyHasAPhoneNumber(Exception): pass # Your account already has a phone number attached to it.

# Adding Phone Number Exceptions
class EmailNotVerified(Exception): pass
class InvalidSMSCode(Exception): pass

# Adding Steam Guard Exceptions
class AlreadyHasMobileSteamGuard(Exception): pass
class UnableToGenerateCorrectCodes(Exception): pass